/*====================================================================================================*
* PINEWOOD DERBY - PWD 4-Lane Track Timer                                  Version 3.1 - 10 Mar. 2022 *
* Created by: Bill van Leeuwen (email: Billv923@outlook.com) for the Arduino UNO microcontroller      *
*                                                                                                     *
* This program is designed to work in conjunction with the GrandPrix Race Manager(tm)(GPRM) software  *
* by Lisano Enterprises, LLC (http://grandprix-software-central.com).                                 *
*                                                                                                     *
* This program measures the start-to-finish time of 4 lanes. Race status and finish times are sent to * 
* the serial port to interface with the GrandPrix Race Manager (GPRM) software running on the PC.     *
* Time is measured from activation of the start switch to activation of the finish line sensor of     *
* each lane. Each car has 10 seconds to finish the race. If a car doesn't cross the finish line       *
* within 10 seconds, it is assumed that the car had a problem or the lane was not used. In either     *
* case, the value 9.9999 is assigned to the lane time as an error indication for use by the GPRM      *
* software running on the PC.                                                                         *
*                                                                                                     *
* This design uses infrared sensors (one for each lane) embedded in the track at the finish line to   *
* detect when a car crosses the finish line. An infrared light source mounted above the finish line   *
* provide the illumination needed by the sensors. A switch connected to the start gate is used to     *
* start the timer.                                                                                    *
*                                                                                                     *
* Refer to the Schematic Diagram, parts list and design notes that accompany this software for wiring * 
* of the infrared lane sensors, race start switch, and reset switch.                                  *
* Feel free to contact me at Billv923@outlook.com if any questions.                                   *
*====================================================================================================*/

//Initialization
#define Num_Lanes    4         //Number of lanes
#define NotReady     0         //Race State = Not Ready
#define Ready        1         //Race State = Ready
#define Racing       2         //Race State = Racing
#define Finished     3         //Race State = Finished

//GPRM Serial In Message Codes: '<-' = Incoming from GPRM
#define Smsg_CkGate  "G"       // <- check gate
#define Smsg_L1Mask  "M1"      // <- mask lane 1
#define Smsg_L2Mask  "M2"      // <- mask lane 2
#define Smsg_L3Mask  "M3"      // <- mask lane 3
#define Smsg_L4Mask  "M4"      // <- mask lane 4
#define Smsg_UnMask  "U"       // <- unmask all lanes
#define Smsg_Reset   "R"       // <- reset
#define Smsg_Test    "T"       // <- Display test character received from PC via Serial port
#define Smsg_Debug   "D"       // <- Debug Request
//#define Smsg_Solen    "S"     // <- start solenoid (Not used)
//#define Smsg_Force    "F"     // <- force end (Not used)
//GPRM Serial Out Message Codes: '->' = Outgoing to GPRM
#define Smsg_Gopen   'O'       // -> gate open
#define Smsg_Ready   'K'       // -> ready
#define Smsg_Start   'B'       // -> race started
#define Smsg_Acknw   '@'       // -> Acknowledge

//Pin Assignments
#define fs1 2                  //Finish Sensor Lane 1 - Digital Input Pin 2
#define fs2 3                  //Finish Sensor Lane 2 - Digital Input Pin 3
#define fs3 4                  //Finish Sensor Lane 3 - Digital Input Pin 4
#define fs4 5                  //Finish Sensor Lane 4 - Digital Input Pin 5
#define gate_sw 8              //Gate Switch (Start)  - Digital Input Pin 8
#define reset_sw 9             //Manual Reset         - Digital Input Pin 9
#define LED 13                 //IO Pin 13 LED        - Digital Pin 13

//Variable Initialization
int i, n;                      //Working variables
int lane_place[Num_Lanes];     //Lane finish place array
boolean track_ready;           //False if any lane sensor is obstructed or Gate Switch Open
boolean Lane1_Mask;            //True if lane 1 is masked
boolean Lane2_Mask;            //True if lane 2 is masked
boolean Lane3_Mask;            //True if lane 3 is masked
boolean Lane4_Mask;            //True if lane 4 is masked
boolean Debug_Err;             //True if Debug routine found a problem
byte regData;                  //Pins D0 - D7 'D' register input data
int Masked_Lanes;              //Number of lanes that are masked
int Active_Lanes;              //Number of active lanes (Total lanes - masked lanes)
int serial_data;               //Serial data
int Racestate;                 //0=NotReady, 1=Ready, 2=Racing, 3=Finished
int fincount;                  //Tallies cars crossing finish line to determine race over
unsigned long Start_Time;      //Race start time (microseconds)
unsigned long Exp_Time;        //Expiration Time (milliseconds)
unsigned long Clk_Time;        //CPU Clock Time (microseconds)
float Time;                    //Working variable - Calculated Lane Time (seconds)
float lane_time[Num_Lanes];    //Calculated lane time data array (seconds)
String Str1;                   //String variable used in serial coms

//======================================= SETUP ROUTINE =======================================
void setup() {  
//Hardware Setup
  // Initialize Finish Sensor, Gate Switch and Reset Switch pins as inputs & pin 13 as an output
  pinMode(fs1, INPUT_PULLUP);                     //Note: Sensor inputs fs1 thru fs4 transition from LOW to
  pinMode(fs2, INPUT_PULLUP);                     //  HIGH when light beam is interrupted (e.g. car
  pinMode(fs3, INPUT_PULLUP);                     //  crosses finish line)
  pinMode(fs4, INPUT_PULLUP);
  pinMode(gate_sw, INPUT_PULLUP);                 //Gate_sw is normally LOW & transitions to HIGH at race start
  pinMode(reset_sw, INPUT_PULLUP);                //Reset_sw is normally HIGH & transitions to LOW when pressed
  pinMode(LED, OUTPUT);                           //set so we can turn off the pin 13 LED

//Software Setup
  Serial.begin(9600);                             //Enable serial comms to the GPRM at 9600 baud
  Serial.setTimeout(10);                          //Limit time to wait for input to 10ms
  Racestate = NotReady;                           //Set initial race state to not ready
  Lane1_Mask = false;                             //Preset Lane 1 mask flag to false
  Lane2_Mask = false;                             //Preset Lane 2 mask flag to false
  Lane3_Mask = false;                             //Preset Lane 3 mask flag to false
  Lane4_Mask = false;                             //Preset Lane 4 mask flag to false
  Masked_Lanes = 0;                               //Preset masked lane count to zero
  digitalWrite(LED,LOW);                          //Turn off Pin 13 LED just for the heck of it
}

//======================================== MAIN LOOP ============================================
void loop() {
  //READ & PROCESS INCOMING SERIAL DATA FROM THE GPRM
  Str1 = get_serial_data();                       //Get any received serial data
  if (Str1.indexOf(Smsg_Reset)>-1) {              //Check for 'Reset' request from GPRM
    Check_Track();                                //If so, call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      Racestate = Ready;                          //Set Racestate to Ready
      Serial.println(Smsg_Ready);                 //Send Ready state message to GPRM
    } 
    else {                                        //Do this if track is not ready
      Racestate = NotReady;                       //Set Racestate to Not Ready
      Serial.println(Smsg_Gopen);                 //Send Not Ready state message to GPRM
    }
  }
  if (Str1.indexOf(Smsg_CkGate)>-1) {             //Check for 'check gate' request from GPRM
    if (digitalRead(gate_sw)==1) {                //Check if Gate Switch is open
      Serial.println(Smsg_Gopen);                 //If so, send Gate Switch not ready message to GPRM
    }
    else {
      Serial.println(Smsg_Acknw);                 //Else send Acknowledge message to GPRM
    }
  }
  if (Str1.indexOf(Smsg_UnMask)>-1) {             //Check for lane Unmask request from GPRM
    Masked_Lanes = 0;                             //Clear masked lane count
    Lane1_Mask = false;                           //Clear Lane 1 mask flag
    Lane2_Mask = false;                           //Clear Lane 2 mask flag
    Lane3_Mask = false;                           //Clear Lane 3 mask flag
    Lane4_Mask = false;                           //Clear Lane 4 mask flag
    Serial.println(Smsg_Acknw);                   //Send Acknowledge message to GPRM
  }
  if (Str1.indexOf(Smsg_L1Mask)>-1) {             //Check for lane 1 mask request from GPRM
    Lane1_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (Str1.indexOf(Smsg_L2Mask)>-1) {             //Check for lane 2 mask request from GPRM
    Lane2_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (Str1.indexOf(Smsg_L3Mask)>-1) {             //Check for lane 3 mask request from GPRM
    Lane3_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (Str1.indexOf(Smsg_L4Mask)>-1) {             //Check for lane 4 mask request from GPRM
    Lane4_Mask = true;                            //Set to TRUE if so
    ++Masked_Lanes;                               //Increment masked lane count
  }
  if (Str1.indexOf(Smsg_Debug)>-1) {              //Check for 'debug' request from PC
    Debug_Err = false;                            //Preset debug error flag to FALSE
    Serial.println("Debug Status:");
    if (digitalRead(gate_sw)==1) {                //Check if Gate Switch is open
      Debug_Err = true;                           //Set debug error flag to TRUE
      Serial.println("- Gate Switch: OPEN");      //Send Gate Switch status to PC
    }
    if (digitalRead(reset_sw)==0) {               //Check if Gate Switch is open
      Debug_Err = true;                           //Set debug error flag to TRUE
      Serial.println("- Reset Switch: CLOSED");   //Send Reset Switch status to PC
    }
    for (n=0; n<Num_Lanes; n++) {                 //Loop through all lanes and
      if (digitalRead(n+2)==1) {                  //Check if a lane sensor is obstructed
        Debug_Err = true;                         //Set debug error flag to TRUE
        Serial.print("- Lane ");                  //Send lane sensor status to PC
        Serial.print(n+1);
        Serial.println(" OBSTRUCTED");
      }
    }
    if (!Debug_Err) Serial.println("- All Inputs OK");  //Send Debug OK Msg to PC
  }
 
  //CHECK & PROCESS RACE STATES
  if (digitalRead(reset_sw)==0){                  //Manual timer Reset button pressed?
    Lane1_Mask = false;                           //Clear Lane 1 mask flag in case it was set
    Lane2_Mask = false;                           //Clear Lane 2 mask flag in case it was set
    Lane3_Mask = false;                           //Clear Lane 3 mask flag in case it was set
    Lane4_Mask = false;                           //Clear Lane 4 mask flag in case it was set
    Masked_Lanes = 0;                             //Clear masked lane count in case it was set
    Check_Track();                                //Call Check Track subroutine
    if (track_ready){                             //and only do a reset if Track is ready
      initialize();                               //Call initialize function & reset everything
      Racestate = Ready;                          //Set Racestate to Ready
      Serial.println(Smsg_Ready);                 //Send Ready state message to GPRM
    } 
    else {                                        //Do this if track is not ready
      Racestate = NotReady;                       //Set Racestate to NotReady
      Serial.println(Smsg_Gopen);                 //Send Not Ready state message to GPRM
    }
    delay(500);                                   //Insert delay to debounce reset switch
  }
  // Transition from Ready to Racing when the Gate switch is tripped
  if (Racestate == Ready && digitalRead(gate_sw)==1) {
    Start_Time = micros();                        //Get start time from CPU clock in microseconds
    Exp_Time = Start_Time + 10000000;             //Set Expiration = Start_Time + 10 seconds (in microseconds)
    Racestate = Racing;                           //Set Racestate to Racing
    Serial.println(Smsg_Start);                   //Send Race start message to GPRM
    delay(200);                                   //Wait a while before we start finish check loop
    
    Finish_Check: //Race underway. Check/Wait for everyone to finish
    /*====================================================================================================*
    * NOTE!!!!: The 'while' loop below has been optimized to repeatedly check lane sensors as fast as     *
    * possible by accessing the digital port input register directly via the PIND command instead of the  *
    * 'digitalRead()' command. This command is specifically tailored to Arduino boards using the ATmega8, *
    * ATmega168 or ATmega328p chips. If this code is used with an Arduino clone having a different        *
    * chip, it may not work.                                                                              *
    *====================================================================================================*/
    Active_Lanes = Num_Lanes-Masked_Lanes;        //Calculate number of active lanes
    while (fincount<Active_Lanes) {               //Loop until all cars have crossed finish line
      Clk_Time = micros();                        //Get the CPU clock time in microseconds
      regData = PIND;                             //Parallel read the D0-D7 data register
      if (lane_place[0]==0) {                     //Check & skip if Lane 1 already finished
        if (!Lane1_Mask) {                        //Check & skip if Lane 1 is masked
          if (regData & 0x04) {                   //Is Car over Lane 1 Finish Line?
            lane_time[0]=Clk_Time;                //Yes...get time
            lane_place[0]=++fincount;             //Incr finish count & set finish place for Lane 1
          }
        }
      }
      if (lane_place[1]==0) {                     //Check & skip if Lane 2 already finished
        if (!Lane2_Mask) {                        //Check & skip if Lane 2 is masked
          if (regData & 0x08) {                   //Is Car over Lane 2 Finish Line?
            lane_time[1]=Clk_Time;                //Yes...get time
            lane_place[1]=++fincount;             //Incr finish count & set finish place for Lane 2
          }
        }
      }  
      if (lane_place[2]==0) {                     //Check & skip if Lane 3 already finished
        if (!Lane3_Mask) {                        //Check & skip if Lane 3 is masked
          if (regData & 0x10) {                   //Is Car over Lane 3 Finish Line?
            lane_time[2]=Clk_Time;                //Yes...get time
            lane_place[2]=++fincount;             //Incr finish count & set finish place for Lane 3
          }
        }
      }
      if (lane_place[3]==0) {                     //Check & skip if Lane 4 already finished
        if (!Lane4_Mask) {                        //Check & skip if Lane 4 is masked
          if (regData & 0x20) {                   //Is Car over Lane 4 Finish Line?
            lane_time[3]=Clk_Time;                //Yes...get time
            lane_place[3]=++fincount;             //Incr finish count & set finish place for Lane 4
          }
        }
      }
      if (Clk_Time>Exp_Time) break;               //Exit 'while' loop if time has expired
    }
    
    //Transition from Racing to Finished. Race is over or time has expired
    Racestate=Finished;                           //Set Race state to Finished
    //Send Race Results To GPRM
    for (int n=0; n<Num_Lanes; n++) {             //Loop thru all lanes
      Serial.print(n+1);                          //Send Lane number
      Serial.print(" ");                          //Add a space after the lane number
      if (lane_time[n]>0) {                       //Check to see if lane time is valid
        Time = (float)((lane_time[n]-Start_Time)/1000000.0);  //Calc lane time and convert to seconds
        Serial.println(Time,4);                   //Send lane time to GPRM trimmed to 4 places after decimal
      }
      else {
        Serial.println("9.9999");                  //Send all nines to GPRM if lane time is invalid
      }
    }
    delay(100);                                   //Wait a while
  }                                               //End Racing "if"
}                                                 //Loop back to void

/*================================================================================*
  SUBROUTINES & FUNCTIONS
 *================================================================================*/
//Subroutine: Initialize Racing Variables
void initialize() {  
  for (int n=0; n<Num_Lanes; n++) {           //Reset lane Place & Times
    lane_place[n] = 0;
    lane_time[n] = 0;
  }
  fincount = 0;                               //Reset finish count to 0
  Serial.flush();                             //Clear the serial input buffer
  return;
}

//Function: Read Serial Data
String get_serial_data() {
  String Sdata = "";                          //Clear any old data
  if (Serial.available() > 0) {               //Check for new data line
    Sdata = Serial.readStringUntil('\n');     //and get it if so
  }
  return Sdata;                               //Return serial data to main program
}

//Function: Check Track - Sets 'track_ready' flag to true or false
int Check_Track() {
  track_ready = true;                         //Default track ready status to true
  if (digitalRead(gate_sw)==1) {              //Check if Gate Switch is open
    track_ready = false;                      //And set track ready status to false
  }
  for (int n=0; n<Num_Lanes; n++) {           //Loop through all lanes and
    if (digitalRead(n+2)==1) {                //Check if a lane sensor (n+2 = pin number) is obstructed
      track_ready = false;                    //If so, set track ready status to false
    }
  }
}
